/*************************************************************************
 *
 *    Used with ICCARM and AARM.
 *
 *    (c) Copyright IAR Systems 2006
 *
 *    File name   : STR91x_usb.c
 *    Description : usb module (HAL)
 *
 *    History :
 *    1. Data        : February 10, 2006
 *       Author      : Stanimir Bonev
 *       Description : Create
 *
 *    $Revision: 1.0 $
 **************************************************************************/

#define HAL_USB_GLOBAL
#include "STR91x_usb.h"

#pragma segment="USB_PACKET_MEMORY"

#pragma location="USB_PACKET_MEMORY"
__root __no_init Int8U USB_PacketMemory[2048];

//pUSB_BuffDeskTbl_t pUSB_BuffDeskTbl;
pInt16U pUSB_BuffDeskTbl;

pPacketMemUse_t pPacketMemUse;

EpInfo_t EpInfo[ENP_MAX_NUMB];
Int32U DlyCnt;

UserFunc_t UsbUserFun [UsbLastEvent] =
{
  // Ep 0 Out
  NULL,
  // Ep 0 In
  NULL,
  // Ep 1 Out
  NULL,
  // Ep 1 In
  NULL,
  // Ep 2 Out
  NULL,
  // Ep 2 Int
  NULL,
  // Ep 3 Out
  NULL,
  // Ep 3 In
  NULL,
  // Ep 4 Out
  NULL,
  // Ep 4 In
  NULL,
  // Ep 5 Out
  NULL,
  // Ep 5 In
  NULL,
  // Ep 6 Out
  NULL,
  // Ep 6 In
  NULL,
  // Ep 7 Out
  NULL,
  // Ep 7 In
  NULL,
  // Ep 8 Out
  NULL,
  // Ep 8 In
  NULL,
  // Ep 9 Out
  NULL,
  // Ep 9 In
  NULL,
  // Ep 10 Out
  NULL,
  // Ep 10 In
  NULL,
  // Ep 11 Out
  NULL,
  // Ep 11 In
  NULL,
  // Ep 12 Out
  NULL,
  // Ep 12 In
  NULL,
  // Ep 13 Out
  NULL,
  // Ep 13 In
  NULL,
  // Ep 14 Out
  NULL,
  // Ep 14 In
  NULL,
  // Ep 15 Out
  NULL,
  // Ep 15 In
  NULL,
  // UsbResetEvent
  NULL,
  // UsbConnectEvent
  NULL,
  // UsbSuspendEvent
  NULL,
  // UsbErrorEvent
  NULL,
  // UsbSofEvent
  NULL,
  // UsbWakeUpEvent
  NULL,
  // UsbDovrEvent
  NULL,
};

/*************************************************************************
 * Function Name: EpCtrlSet_EA
 * Parameters: volatile pInt32U, Int32U Data
 *
 * Return: none
 *
 * Description: Set EP Addr
 *
 *************************************************************************/
void EpCtrlSet_EA(volatile pInt32U pReg, Int32U Data)
{
  Data &= 0xF;
  Data |= *pReg & 0x0700 ;
  Data |= USB_CTRL_CLEAR_ONLY_MASK;
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_EA
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int16U
 *
 * Description: Get EP Addr
 *
 *************************************************************************/
inline
Int16U EpCtrlGet_EA(volatile pInt32U pReg)
{
  return(*pReg & 0x0F);
}

/*************************************************************************
 * Function Name: EpCtrlSet_STAT_TX
 * Parameters: volatile pInt32U pReg, Int32U Data
 *
 * Return: none
 *
 * Description: Set EP IN State
 *
 *************************************************************************/
void EpCtrlSet_STAT_TX (volatile pInt32U pReg, Int32U Data)
{
  Data &= 0x3;
  Data <<= 4;
  Data ^= *pReg;
  Data |= USB_CTRL_CLEAR_ONLY_MASK;
  Data &= ~0x7040;  // Clear the toggle bits without STAT_TX (4,5)
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_STAT_TX
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP IN Status
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_STAT_TX (volatile pInt32U pReg)
{
Int32U Data = *pReg;
  return((Data & (0x3UL << 4)) >> 4);
}

/*************************************************************************
 * Function Name: EpCtrlSet_DTOG_TX
 * Parameters: volatile pInt32U pReg, Int32U Data
 *
 * Return: none
 *
 * Description: Set EP DTOG_TX bit (IN)
 *
 *************************************************************************/
void EpCtrlSet_DTOG_TX(volatile pInt32U pReg, Int32U Data)
{
  Data  = Data?(*pReg ^ (1UL<<6)):*pReg;
  Data &= ~0x7030;   // Clear the toggle bits without DTOG_TX (6)
  Data |= USB_CTRL_CLEAR_ONLY_MASK;
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlToggle_DTOG_TX
 * Parameters: volatile pInt32U pReg
 *
 * Return: none
 *
 * Description: Toggle EP DTOG_TX bit (IN)
 *
 *************************************************************************/
void EpCtrlToggle_DTOG_TX(volatile pInt32U pReg)
{
Int32U Data = *pReg | (1UL << 6);
  Data &= ~0x7030;   // Clear the toggle bits without DTOG_TX (6)
  Data |= USB_CTRL_CLEAR_ONLY_MASK;
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_DTOG_TX
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP DTOG_TX bit (IN)
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_DTOG_TX(volatile pInt32U pReg)
{
  return(((*pReg & (1UL<<6))?1:0));
}

/*************************************************************************
 * Function Name: EpCtrlClr_CTR_TX
 * Parameters: volatile pInt32U pReg
 *
 * Return: none
 *
 * Description: Clear EP CTR_TX bit (IN)
 *
 *************************************************************************/
void EpCtrlClr_CTR_TX(volatile pInt32U pReg)
{
Int32U Data = *pReg;
  Data &= ~(USB_CTRL_TOGGLE_MASK | 1UL << 7);
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_CTR_TX
 * Parameters: volatile pInt32U pReg
 *
 * Return: UserFunc_t
 *
 * Description: Get EP CTR_TX bit (IN)
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_CTR_TX(volatile pInt32U pReg)
{
  return(((*pReg & (0x1UL << 7))?1:0));
}

/*************************************************************************
 * Function Name: EpCtrlSet_EP_KIND
 * Parameters: volatile pInt32U pReg, Int32U Data
 *
 * Return: none
 *
 * Description: Set EP KIND bit
 *
 *************************************************************************/
void EpCtrlSet_EP_KIND (volatile pInt32U pReg, Int32U Data)
{
  Data  = Data ? (1UL << 8):0;
  Data |= *pReg & ~(USB_CTRL_TOGGLE_MASK | (1UL << 8));
  Data |= USB_CTRL_CLEAR_ONLY_MASK;
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_EP_KIND
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP KIND bit
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_EP_KIND(volatile pInt32U pReg)
{
  return(*pReg & (0x1UL << 8)?1:0);
}

/*************************************************************************
 * Function Name: EpCtrlSet_EP_TYPE
 * Parameters: volatile pInt32U pReg, Int32U Data
 *
 * Return: none
 *
 * Description: Set EP type
 *
 *************************************************************************/
void EpCtrlSet_EP_TYPE(volatile pInt32U pReg, Int32U Data)
{
  Data  &= 0x3;
  Data <<= 9;
  Data  |= *pReg & ~(USB_CTRL_TOGGLE_MASK | (0x3UL << 9));
  Data  |= USB_CTRL_CLEAR_ONLY_MASK;
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_EP_TYPE
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP type
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_EP_TYPE(volatile pInt32U pReg)
{
Int32U Data = *pReg;
  Data &= (3UL<<9);
  return(Data >> 9);
}

/*************************************************************************
 * Function Name: EpCtrlGet_SETUP
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP SETUP bit
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_SETUP(volatile pInt32U pReg)
{
Int32U Data = *pReg & (0x1UL << 11);
  return(Data?1:0);
}

/*************************************************************************
 * Function Name: EpCtrlSet_STAT_RX
 * Parameters: volatile pInt32U pReg, Int32U Data
 *
 * Return: none
 *
 * Description: Set EP STAT_RX (OUT)
 *
 *************************************************************************/
void EpCtrlSet_STAT_RX (volatile pInt32U pReg, Int32U Data)
{
  Data  &= 0x3;
  Data <<= 12;
  Data  ^= *pReg;
  Data  |= USB_CTRL_CLEAR_ONLY_MASK;
  Data &= ~0x4070;  // Clear the toggle bits without STAT_RX (12,13)
  *pReg  = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_STAT_RX
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP STAT_RX (OUT)
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_STAT_RX(volatile pInt32U pReg)
{
Int32U Data = *pReg &( 0x3UL << 12);
  return(Data >> 12);
}

/*************************************************************************
 * Function Name: EpCtrlSet_DTOG_RX
 * Parameters: volatile pInt32U pReg, Int32U Data
 *
 * Return: none
 *
 * Description: Set DTOG_RX bit
 *
 *************************************************************************/
void EpCtrlSet_DTOG_RX(volatile pInt32U pReg, Int32U Data)
{
  Data  = Data?(*pReg ^ (1UL<<14)):*pReg;
  Data &= ~0x3070;   // Clear the toggle bits without DTOG_RX (14)
  Data |= USB_CTRL_CLEAR_ONLY_MASK;
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlToggle_DTOG_RX
 * Parameters: volatile Int16U pReg
 *
 * Return: none
 *
 * Description: Toggle DTOG_RX bit
 *
 *************************************************************************/
void EpCtrlToggle_DTOG_RX(volatile pInt32U pReg)
{
Int32U Data = *pReg | (1UL << 14);
  Data &= ~0x3070;   // Clear the toggle bits without DTOG_TX (14)
  Data |= USB_CTRL_CLEAR_ONLY_MASK;
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_DTOG_RX
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP DTOG_RX bit (OUT)
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_DTOG_RX(volatile pInt32U pReg)
{
  return(((*pReg & (1UL<<14))?1:0));
}

/*************************************************************************
 * Function Name: EpCtrlClr_CTR_RX
 * Parameters: volatile pInt32U pReg
 *
 * Return: none
 *
 * Description: Clear EP CTR_RX bit (OUT)
 *
 *************************************************************************/
void EpCtrlClr_CTR_RX(volatile pInt32U pReg)
{
Int32U Data = *pReg;
  Data &= ~(USB_CTRL_TOGGLE_MASK | 1UL << 15);
  *pReg = Data;
}

/*************************************************************************
 * Function Name: EpCtrlGet_CTR_RX
 * Parameters: volatile pInt32U pReg
 *
 * Return: Int32U
 *
 * Description: Get EP CTR_RX bit (OUT)
 *
 *************************************************************************/
inline
Int32U EpCtrlGet_CTR_RX(volatile pInt32U pReg)
{
  return(*pReg & (0x1UL << 15)?1:0);
}

/*************************************************************************
 * Function Name: WriteEpDTB
 * Parameters: EpSlot_t Slot, Int32U Offset, Int32U Data
 *
 * Return: none
 *
 * Description: Write to EP Buffer descriptor table
 *
 *************************************************************************/
void WriteEpDTB (EpSlot_t Slot, Int32U Offset, Int32U Data)
{
pInt32U pUSB_BuffDeskTbltt;
Int32U DescTableHold;
  pUSB_BuffDeskTbltt  = (pInt32U)((Int32U)__segment_begin("USB_PACKET_MEMORY")+(Int32U)USB_OffsetOfDBT);
  pUSB_BuffDeskTbltt += (sizeof(USB_BuffDeskTbl_t)/sizeof(Int32U)) * Slot;

  switch (Offset)
  {
  case 0:
    DescTableHold  = *pUSB_BuffDeskTbltt & ~0xFFFF;
    DescTableHold |= Data & 0xFFFF;
    *pUSB_BuffDeskTbltt = DescTableHold;
    break;
  case 2:
    DescTableHold  = *pUSB_BuffDeskTbltt & 0xFFFF;
    DescTableHold |= (Data & 0xFFFF) << 16;
    *pUSB_BuffDeskTbltt = DescTableHold;
    break;
  case 4:
    DescTableHold  = *(pUSB_BuffDeskTbltt+1) & ~0xFFFF;
    DescTableHold |= Data & 0xFFFF;
    *(pUSB_BuffDeskTbltt+1) = DescTableHold;
    break;
  case 6:
    DescTableHold  = *(pUSB_BuffDeskTbltt+1) & 0xFFFF;
    DescTableHold |= (Data & 0xFFFF) << 16;
    *(pUSB_BuffDeskTbltt+1) = DescTableHold;
    break;
  }
}

/*************************************************************************
 * Function Name: USB_UserFuncRegistered
 * Parameters: UserFunc_t UserFunc, UsbUserEvent_t UserFuncInd
 *
 * Return: UserFunc_t
 *
 * Description: Registered User callback function
 *
 *************************************************************************/
UserFunc_t USB_UserFuncRegistered (UserFunc_t UserFunc, UsbUserEvent_t UserFuncInd)
{
UserFunc_t PrevUserFunc = UsbUserFun[UserFuncInd];
  UsbUserFun[UserFuncInd] = (UserFunc_t) UserFunc;
  return (PrevUserFunc);
}

/*************************************************************************
 * Function Name: USB_HwInit
 * Parameters: VicIrqSlots_t IntrSlotHi, VicIrqSlots_t IntrSlotLo
 *
 * Return: none
 *
 * Description: Init USB
 *
 *************************************************************************/
void USB_HwInit(VicIrqSlots_t IntrSlotHi, VicIrqSlots_t IntrSlotLo)
{
  // Init P27 as USB Clock
  SCU_GPIOOUT2_bit.P2OUT7   = 0;
  SCU_GPIOIN2_bit.P2IN7     = 0;
  // Init P70 GPIO Out
//  SCU_PCGR1_bit.GPIO7       = 1;  // GPIO7 Port clock runing
  SCU_PRR1_bit.RST_GPIO7    = 1;  // GPIO7 out from reset
  SCU_GPIOTYPE7_bit.P7TYPE0 = 0;  // GPIO7.0 Push-Pull outputs
  SCU_GPIOIN7_bit.P7IN0     = 0;  // GPIO7.0 On-chip peripheral not connected to the input
  SCU_GPIOOUT7_bit.P7OUT0   = 1;  // GPIO7.0 Alternate Output 1 (general purpose output)
  GPIO7_DATA_bit.no0        = 1;
  GPIO7_DIR_bit.no0         = 1;
  GPIO7_SEL_bit.no0         = 0;  // GPIO7.0 assign to GPIO

  // Reset USB Engine
  SCU_PRR0_bit.RST_USB = 0;
  // Init USB Clock
  // Enable clock
  SCU_PCGR0_bit.USB = 1;
  // Release reset USB Engine
  SCU_PRR0_bit.RST_USB = 1;

  // Exit from power down
  USB_CNTR_bit.PDWN = 0;
  // Force USB Reset
  USB_CNTR_bit.FRES = 1;
  // Clear USB reset &  Disable USB interrups
  USB_CNTR = 0;

  // Clear spurious pending interrupt
  USB_ISTR = 0;

  // Select ext. 48MHz clock
//  SCU_CLKCNTR_bit.USBSEL = 2;
  // Select Fmstr/2
  SCU_CLKCNTR_bit.USBSEL = 1;
  // Enable USB48M clock
  SCU_PCGR0_bit.USB48M = 1;

  // USB interrupt connect to VIC
#if USB_HIGH_PRIORITY_EVENT > 0
  VIC_Config(USBHP_ITLine,VIC_IRQ , IntrSlotHi);
  VIC_ITCmd(USBHP_ITLine, ENABLE);
  VIC_Config(USBLP_ITLine,VIC_IRQ , IntrSlotHi);
  VIC_ITCmd(USBLP_ITLine, ENABLE);
#else
  VIC_Config(USBLP_ITLine,VIC_IRQ , IntrSlotLo);
  VIC_ITCmd(USBLP_ITLine, ENABLE);

#endif

  // Disconnect device
  USB_Connect(FALSE);
  // Init controls endpoints
  USB_HwReset();

  // Enable Device interrups
  USB_CNTR = bmCTRM | bmRESETM | bmSUSPM |
         (USB_WKUP_EVENT  ? bmWKUPM : 0) |
         (USB_SOF_EVENT   ? bmSOFM  : 0) |
         (USB_ERROR_EVENT ? bmERRM  : 0) |
         ( USB_DOVR_EVENT ? bmDOVRM : 0) ;
}

/*************************************************************************
 * Function Name: USB_HwReset
 * Parameters: none
 *+
 * Return: none
 *
 * Description: Reset Usb engine
 *
 *************************************************************************/
void USB_HwReset (void)
{
Int32U Count;
  for(Count = 0; Count < ENP_MAX_NUMB; Count++)
  {
    EpInfo[Count].pEpCtrl = NULL;
  }

  for (Count = 0; Count < sizeof(USB_PacketMemory)/sizeof(USB_PacketMemory[0]); Count++)
  {
    USB_PacketMemory[Count] = 0xCC;
  }

#if USB_DMA > 0
  // DMA transfer disable
  USB_DmaReset();
#endif
  // Resume delay counter
  DlyCnt = 0;
  // Clear USB Reset
  USB_CNTR_bit.FRES = 0;
  // USB_Configure
  USB_Configure(FALSE);
  // Set Buffer table address
  USB_BTABLE = USB_OffsetOfDBT;
  pUSB_BuffDeskTbl = (pInt16U)((Int32U)__segment_begin("USB_PACKET_MEMORY") + USB_OffsetOfDBT);
  // Init Packet memory table
  pPacketMemUse = NULL;
  // EndPoint Init
  USB_RealizeEndPoint(CTRL_ENP_OUT,0,Ep0MaxSize,TRUE,EP_CTRL,EP_SLOT0);
  USB_RealizeEndPoint(CTRL_ENP_IN, 1,Ep0MaxSize,TRUE,EP_CTRL,EP_SLOT0);
  // Set Arrdess 0
  USB_SetDefAdd();
}

/*************************************************************************
 * Function Name: USB_AllocateBuffer
 * Parameters: pInt16U pOffset,Int32U PacketSize,
 *             USB_Endpoint_t EndPoint
 *
 * Return: Boolean
 *
 * Description: Allocate a new buffer
 *
 *************************************************************************/
Boolean USB_AllocateBuffer (pInt16U pOffset,pInt32U pPacketSize,
                            USB_Endpoint_t EndPoint)
{
pPacketMemUse_t  pPacketMem = pPacketMemUse, pPacketMemNext, pPacketMemUseNew;
//Int32U MaxPacketSize = *pPacketSize+2;  // for CRC
Int32U MaxPacketSize = *pPacketSize;
  if(MaxPacketSize > 124)
  {
    // coarse-granularity size alignment by 32
    MaxPacketSize +=  (32-1);
    MaxPacketSize &= ~(32-1);
  }
  else
  {
    // fine-granularity size alignment by 2
    MaxPacketSize +=  (2-1);
    MaxPacketSize &= ~(2-1);
  }
  *pOffset = 0;
  while(pPacketMem != NULL)
  {
    *pOffset = pPacketMem->Start + pPacketMem->Size;
    // Offset alignment by 4
    *pOffset += (4-1);
    *pOffset &= ~(4-1);
    pPacketMemNext = pPacketMem->pNext;
    if((pPacketMem->pNext == NULL) ||
       (pPacketMemNext->Start >= *pOffset + MaxPacketSize))
    {
      break;
    }
    pPacketMem = pPacketMem->pNext;
  }
  if((*pOffset + MaxPacketSize) >=
     (((Int32U)__segment_size("USB_PACKET_MEMORY") -
       sizeof(USB_BuffDeskTbl_t) * (MaxIndOfRealizeEp+1))&~7))
  {
    return(FALSE);
  }
  pPacketMemUseNew = malloc(sizeof(PacketMemUse_t));
  if(pPacketMemUseNew == NULL)
  {
    return(FALSE);
  }
  if(pPacketMemUse == NULL)
  {
    pPacketMemUse = pPacketMemUseNew;
    pPacketMemUse->pNext = NULL;
  }
  else
  {
    pPacketMemUseNew->pNext  = pPacketMem->pNext;
    pPacketMem->pNext        = pPacketMemUseNew;
  }
  pPacketMemUseNew->RpAddr = EndPoint;
  pPacketMemUseNew->Start  = *pOffset;
  pPacketMemUseNew->Size   = MaxPacketSize;
  *pPacketSize = MaxPacketSize;
  return(TRUE);
}

/*************************************************************************
 * Function Name: USB_ReleaseBuffer
 * Parameters: USB_Endpoint_t EndPoint
 *
 * Return: none
 *
 * Description: Release the buffer/s
 *
 *************************************************************************/
void USB_ReleaseBuffer (USB_Endpoint_t EndPoint)
{
pPacketMemUse_t  pPacketMem, pPacketMemPrev;
  pPacketMem = pPacketMemUse;
  while(pPacketMem != NULL)
  {
    if(pPacketMem->RpAddr == EndPoint)
    {
      pPacketMemPrev->pNext = pPacketMem->pNext;
      free(pPacketMem);
      pPacketMem = pPacketMemPrev->pNext;
    }
    else
    {
      pPacketMemPrev = pPacketMem;
      pPacketMem = pPacketMem->pNext;
    }
    if (pPacketMem == NULL)
    {
      break;
    }
  }
}

/*************************************************************************
 * Function Name: USB_RealizeEndPoint
 * Parameters: USB_Endpoint_t EndPoint, Boolean EpPrio,
 *             Int32U MaxPacketSize, Boolean IntrEna,
 *             EpType_t EpKind, EpSlot_t EpSlot,
 *
 * Return: USB_ErrorCodes_t
 *
 * Description: Enable or disable endpoint
 *
 *************************************************************************/
USB_ErrorCodes_t USB_RealizeEndPoint( USB_Endpoint_t EndPoint, Boolean EpPrio,
                                      Int32U MaxPacketSize, Boolean IntrEna,
                                      EpType_t EpType, EpSlot_t EpSlot)
{
volatile pInt32U pEpCtrlBaseAddr;
Int16U Offset,Offset1;
Int32U MaxPacketSizeTmp;
  if (MaxPacketSize)
  {
    // Allocate packet memory for EP buffer/s
    // calculate actual size only for the OUT EPs
    MaxPacketSizeTmp = MaxPacketSize;
    if(!USB_AllocateBuffer(&Offset,&MaxPacketSizeTmp,EndPoint))
    {
      return(USB_MEMORY_FULL);
    }
    if((EpType == EP_ISO) || (EpType == EP_BULK_DOUB_BUF))
    {
      // Allocate packet second buffer
      MaxPacketSizeTmp = MaxPacketSize;
      if(!USB_AllocateBuffer(&Offset1,&MaxPacketSizeTmp,EndPoint))
      {
        // release buffer
        USB_ReleaseBuffer(EndPoint);
        return(USB_MEMORY_FULL);
      }
    }

    pEpCtrlBaseAddr  = (pInt32U)&USB_EP0R;
    pEpCtrlBaseAddr += EpSlot;
    // Set Ep Address
    EpCtrlSet_EA(pEpCtrlBaseAddr,EndPoint >> 1);
    EpInfo[EndPoint].pEpCtrl = pEpCtrlBaseAddr;
    EpInfo[EndPoint].EpSlot  = EpSlot;
    EpInfo[EndPoint].MaxSize = MaxPacketSize;
    if(EndPoint & 1)
    {
      // IN EP
      // Disable EP
      EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_DISABLED);
      // Clear Tx toggle
      EpCtrlSet_DTOG_TX(pEpCtrlBaseAddr,0);
      // Clear Correct Transfer for transmission flag
      EpCtrlClr_CTR_TX(pEpCtrlBaseAddr);
      // Update EP description table
      WriteEpDTB_AddrTx(EpSlot,Offset);
      WriteEpDTB_CountTx(EpSlot,0);
      // Set EP Kind & enable
      switch(EpType)
      {
      case EP_BULK_SIG_BUFF:
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_BULK_SIG_BUFF);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        // Enable EP
        EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_NAK);
        break;
      case EP_CTRL:
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_CTRL);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        // Enable EP
        EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_NAK);
        break;
      case EP_ISO:
        // Disable receiving (only singe direction is possible)
        EpCtrlSet_STAT_RX(pEpCtrlBaseAddr,0);
        WriteEpDTB_AddrRx(EpSlot,Offset1);
        WriteEpDTB_CountRx(EpSlot,0);
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_ISO);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        // Enable EP
        EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_VALID);
        break;
      case EP_INTERRUPT:
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_INTERRUPT);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        // Enable EP
        EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_NAK);
        break;
      default:  // EP_BULK_DOUB_BUF
        // Disable receiving (only singe direction is possible)
        EpCtrlSet_STAT_RX(pEpCtrlBaseAddr,EP_DISABLED);
        // Clear Tx Software toggle
        EpCtrlSet_DTOG_RX(pEpCtrlBaseAddr,0);
        WriteEpDTB_AddrRx(EpSlot,Offset1);
        WriteEpDTB_CountRx(EpSlot,0);
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_BULK_SIG_BUFF);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,1);
        // Enable EP
        EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_NAK);
      }
    }
    else
    {
      RxCount_t RxCount = {0};
      // OUT EP
      // Disable EP
      EpCtrlSet_STAT_RX(pEpCtrlBaseAddr,EP_DISABLED);
      // Clear Correct Transfer for reception flag
      EpCtrlClr_CTR_RX(pEpCtrlBaseAddr);
      // Clear Rx toggle
      EpCtrlSet_DTOG_RX(pEpCtrlBaseAddr,0);
      // Update EP description table
      RxCount.BlSizeField = (MaxPacketSize > 124);
      RxCount.NubBlockField = (MaxPacketSize > 124) ? (MaxPacketSize>>5)-1:MaxPacketSizeTmp>>1;
      WriteEpDTB_AddrRx(EpSlot,Offset);
      WriteEpDTB_CountRx(EpSlot,RxCount.Count);

      // Set EP Kind & enable
      switch(EpType)
      {
      case EP_BULK_SIG_BUFF:
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_BULK_SIG_BUFF);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        break;
      case EP_CTRL:
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_CTRL);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        break;
      case EP_ISO:
        WriteEpDTB_CountTx(EpSlot,RxCount.Count);
        // Disable transmitting (only singe direction is possible)
        EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_DISABLED);
        WriteEpDTB_AddrTx(EpSlot,Offset1);
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_ISO);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        break;
      case EP_INTERRUPT:
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_INTERRUPT);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,0);
        break;
      default:  // EP_BULK_DOUB_BUF
        // Disable transmitting (only singe direction is possible)
        EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_DISABLED);
        WriteEpDTB_CountTx(EpSlot,RxCount.Count);
        WriteEpDTB_AddrTx(EpSlot,Offset1);
        EpCtrlSet_EP_TYPE(pEpCtrlBaseAddr,EP_BULK_SIG_BUFF);
        EpCtrlSet_EP_KIND(pEpCtrlBaseAddr,1);
        EpCtrlSet_DTOG_TX(pEpCtrlBaseAddr,0);
      }
      // Enable EP
      EpCtrlSet_STAT_RX(pEpCtrlBaseAddr,EP_VALID);
    }
  }
  else
  {
    pEpCtrlBaseAddr  = (pInt32U)&USB_EP0R;
    pEpCtrlBaseAddr += EpSlot;
    if(EndPoint & 1)
    {
      // Disable IN EP
      EpCtrlSet_STAT_TX(pEpCtrlBaseAddr,EP_DISABLED);
    }
    else
    {
      // Disable OUT EP
      EpCtrlSet_STAT_RX(pEpCtrlBaseAddr,EP_DISABLED);
    }
    // release buffer
    USB_ReleaseBuffer(EndPoint);
    EpInfo[EndPoint].pEpCtrl = NULL;
  }
  return(USB_OK);
}

/*************************************************************************
 * Function Name: USB_SetAdd
 * Parameters: Int32U DevAdd - device address between 0 - 127
 *
 * Return: none
 *
 * Description: Set device address
 *
 *************************************************************************/
void USB_SetAdd(Int32U DevAdd)
{
  USB_DADDR = DevAdd | 0x80;
}

/*************************************************************************
 * Function Name: USB_Connect
 * Parameters: Boolean Conn
 *
 * Return: none
 *
 * Description: Connect Usb
 *
 *************************************************************************/
void USB_Connect (Boolean Conn)
{
  GPIO7_DATA_bit.no0 = !Conn;
}

/*************************************************************************
 * Function Name: USB_Configure
 * Parameters: Boolean Configure
 *
 * Return: none
 *
 * Description: Configure device
 * 							When Configure != 0 enable all Realize Ep
 *
 *************************************************************************/
void USB_Configure (Boolean Configure) {}

/*************************************************************************
 * Function Name: USB_WakeUp
 * Parameters: none
 *
 * Return: none
 *
 * Description: Wake up Usb
 *
 *************************************************************************/
void USB_WakeUp (void)
{
  // Wake-Up
  USB_Resume(USB_RESUME_SOFT_EVENT);
}

/*************************************************************************
 * Function Name: USB_Resume
 * Parameters: UsbResumeEvent_t UsbResumeEvent
 *
 * Return: none
 *
 * Description: USB Resume implement
 *
 *************************************************************************/
void USB_Resume (UsbResumeEvent_t UsbResumeEvent)
{
static Boolean bESOF_IntrEna;
  switch(UsbResumeEvent)
  {
  case USB_RESUME_SOF_EVENT:
    if(DlyCnt)
    {
      if(--DlyCnt == 0)
      {
        USB_CNTR_bit.RESUME = 0;
        USB_CNTR_bit.ESOFM = bESOF_IntrEna;
        if(UsbUserFun[UsbSuspendEvent] != NULL)
        {
          UsbUserFun[UsbSuspendEvent]((void *)0);
        }
      }
    }
    break;
  case USB_RESUME_SOFT_EVENT:
  case USB_RESUME_WAKE_UP_EVENT:
    bESOF_IntrEna = USB_CNTR_bit.SOFM;
    // Eanble SOF
    USB_CNTR_bit.ESOFM = 1;
    DlyCnt = USB_RESUME_DLY;
    USB_CNTR_bit.RESUME = 1;
    break;
  }
}

/*************************************************************************
 * Function Name: USB_GetDevStatus
 * Parameters: USB_DevStatusReqType_t Type
 *
 * Return: Boolean
 *
 * Description: Return USB device status
 *
 *************************************************************************/
Boolean USB_GetDevStatus (USB_DevStatusReqType_t Type)
{
  switch (Type)
  {
  case USB_DevConnectStatus:
    return(TRUE);
  case USB_SuspendStatus:
    return(USB_CNTR_bit.FSUSP);
  case USB_ResetStatus:
    return(USB_CNTR_bit.FRES);
  }
  return(FALSE);
}

/*************************************************************************
 * Function Name: USB_SetStallEP
 * Parameters: USB_Endpoint_t EndPoint, Boolean Stall
 *
 * Return: USB_ErrorCodes_t
 *
 * Description: The endpoint stall/unstall
 *
 *************************************************************************/
USB_ErrorCodes_t USB_SetStallEP (USB_Endpoint_t EndPoint, Boolean Stall)
{
  if( EpInfo[EndPoint].pEpCtrl == NULL)
  {
    return(USB_EP_NOT_VALID);
  }
  if(Stall)
  {
    if(EndPoint & 1)
    {
      // IN EP
      EpCtrlSet_STAT_TX(EpInfo[EndPoint].pEpCtrl,EP_STALL);
    }
    else
    {
      // OUT EP
      EpCtrlSet_STAT_RX(EpInfo[EndPoint].pEpCtrl,EP_STALL);
    }
  }
  else if(EndPoint & 1)
  {
    // IN EP
    EpCtrlSet_STAT_TX(EpInfo[EndPoint].pEpCtrl,EP_NAK);
  }
  else
  {
    // OUT EP
    EpCtrlSet_STAT_RX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
  }
  return(USB_OK);
}

/*************************************************************************
 * Function Name: USB_GetStallEP
 * Parameters: USB_Endpoint_t EndPoint, pBoolean pStall
 *
 * Return: USB_ErrorCodes_t
 *
 * Description: Get stall state of the endpoint
 *
 *************************************************************************/
USB_ErrorCodes_t USB_GetStallEP (USB_Endpoint_t EndPoint, pBoolean pStall)
{
  if(EpInfo[EndPoint].pEpCtrl == NULL)
  {
    return(USB_EP_NOT_VALID);
  }
  *pStall = (EndPoint & 1)?
            (EpCtrlGet_STAT_TX(EpInfo[EndPoint].pEpCtrl) == EP_STALL):  // IN EP
            (EpCtrlGet_STAT_RX(EpInfo[EndPoint].pEpCtrl) == EP_STALL);  // OUT EP
  return (USB_OK);
}

/*************************************************************************
 * Function Name: USB_EpWrite
 * Parameters: USB_Endpoint_t EndPoint, pInt32U pData, pInt32U pCount
 *
 * Return: USB_ErrorCodes_t
 *
 * Description: Endpoint Write (IN)
 *
 *************************************************************************/
USB_ErrorCodes_t USB_EpWrite (USB_Endpoint_t EndPoint, pInt32U pData,
                              pInt32U pCount)
{
EpState_t EpState;
pInt32U pDst;
Int32S  ActSize = *pCount;
Boolean EpOverFlow = FALSE;
  if(EpInfo[EndPoint].pEpCtrl == NULL)
  {
    return(USB_EP_NOT_VALID);
  }

  if(*pCount > EpInfo[EndPoint].MaxSize)
  {
    EpOverFlow = TRUE;
    ActSize    = EpInfo[EndPoint].MaxSize;
  }
  EpState = (EpState_t) EpCtrlGet_STAT_TX(EpInfo[EndPoint].pEpCtrl);

  switch(EpCtrlGet_EP_TYPE(EpInfo[EndPoint].pEpCtrl))
  {
  case EP_INTERRUPT:
    if(EpState == EP_DISABLED)
    {
      // Fatal error (require restart of the USB)
      return(USB_EP_FATAL_ERROR);
    }
    else if (EpState == EP_STALL)
    {
      // EP stalled
      return(USB_EP_STALLED);
    }
    else if(EpState == EP_VALID)
    {
      // Data in progress
      return(USB_EP_OVERRUN);
    }
    WriteEpDTB_CountTx(EpInfo[EndPoint].EpSlot,ActSize);
    pDst = (pInt32U)(ReadEpDTB_AddrTx(EpInfo[EndPoint].EpSlot) +\
                     (Int32U)__segment_begin("USB_PACKET_MEMORY"));
    while(ActSize > 0)
    {
      *pDst++ = *pData++;
      ActSize -= 4;
    }
    // Validate buffer
    EpCtrlSet_STAT_TX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
    break;
  case EP_CTRL:
    if(EpState == EP_VALID)
    {
      return(USB_EP_OVERRUN);
    }
    WriteEpDTB_CountTx(EpInfo[EndPoint].EpSlot,ActSize);
    pDst = (pInt32U)(ReadEpDTB_AddrTx(EpInfo[EndPoint].EpSlot) +\
                     (Int32U)__segment_begin("USB_PACKET_MEMORY"));
    while(ActSize > 0)
    {
      *pDst++ = *pData++;
      ActSize -= 4;
    }
    // Validate buffer
    EpCtrlSet_STAT_TX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
    break;
  case EP_ISO:

    if(EpCtrlGet_DTOG_TX(EpInfo[EndPoint].pEpCtrl))
    {
      // Tx Buffer
      pDst = (pInt32U)(ReadEpDTB_AddrTx(EpInfo[EndPoint].EpSlot) +\
                      (Int32U)__segment_begin("USB_PACKET_MEMORY"));
      WriteEpDTB_CountTx(EpInfo[EndPoint].EpSlot,ActSize);
    }
    else
    {
      // Rx Buffer
      pDst = (pInt32U)(ReadEpDTB_AddrRx(EpInfo[EndPoint].EpSlot) +\
                      (Int32U)__segment_begin("USB_PACKET_MEMORY"));
      WriteEpDTB_CountRx(EpInfo[EndPoint].EpSlot,ActSize);
    }
    while(ActSize > 0)
    {
      *pDst++ = *pData++;
      ActSize -= 4;
    }
  default : // BULK EP
    if(EpCtrlGet_EP_KIND(EpInfo[EndPoint].pEpCtrl))
    {
      Int32U Count;
      // Double buffered
      // Single buffer
      if(EpState == EP_DISABLED)
      {
        // Fatal error (require restart of the USB)
        return(USB_EP_FATAL_ERROR);
      }
      else if (EpState == EP_STALL)
      {
        // EP stalled
        return(USB_EP_STALLED);
      }
      else if(EpState == EP_VALID)
      { // suppress warning
        Boolean DtogHold = EpCtrlGet_DTOG_TX(EpInfo[EndPoint].pEpCtrl);
        if(DtogHold == EpCtrlGet_DTOG_RX(EpInfo[EndPoint].pEpCtrl))
        {
          // Data in progress
          return(USB_EP_OVERRUN);
        }
      }

      if(EpCtrlGet_DTOG_RX(EpInfo[EndPoint].pEpCtrl))
      {
        // Tx Buffer
        pDst = (pInt32U)(ReadEpDTB_AddrRx(EpInfo[EndPoint].EpSlot) +\
                         (Int32U)__segment_begin("USB_PACKET_MEMORY"));
        for(Count = 0; Count < ActSize; Count += 4)
        {
          *pDst++ = *pData++;
        }
        WriteEpDTB_CountRx(EpInfo[EndPoint].EpSlot,ActSize);
      }
      else
      {
        // Rx Buffer
        pDst = (pInt32U)(ReadEpDTB_AddrTx(EpInfo[EndPoint].EpSlot) +\
                         (Int32U)__segment_begin("USB_PACKET_MEMORY"));
        for(Count = 0; Count < ActSize; Count += 4)
        {
          *pDst++ = *pData++;
        }
        WriteEpDTB_CountTx(EpInfo[EndPoint].EpSlot,ActSize);
      }
      EpCtrlToggle_DTOG_RX(EpInfo[EndPoint].pEpCtrl);
      if(EpCtrlGet_STAT_TX(EpInfo[EndPoint].pEpCtrl) != EP_VALID)
      {
        EpCtrlSet_STAT_TX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
      }
    }
    else
    {
      // Single buffer
      if(EpState == EP_DISABLED)
      {
        // Fatal error (require restart of the USB)
        return(USB_EP_FATAL_ERROR);
      }
      else if (EpState == EP_STALL)
      {
        // EP stalled
        return(USB_EP_STALLED);
      }
      else if(EpState == EP_VALID)
      {
        // Data in progress
        return(USB_EP_OVERRUN);
      }
      WriteEpDTB_CountTx(EpInfo[EndPoint].EpSlot,ActSize);
      pDst = (pInt32U)(ReadEpDTB_AddrTx(EpInfo[EndPoint].EpSlot) +\
                       (Int32U)__segment_begin("USB_PACKET_MEMORY"));
      while(ActSize > 0)
      {
        *pDst++ = *pData++;
        ActSize -= 4;
      }
      // Validate buffer
      EpCtrlSet_STAT_TX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
    }
  }

  if(EpOverFlow)
  {
    *pCount = EpInfo[EndPoint].MaxSize;
    return(USB_BUF_OVERFLOW);
  }
  return(USB_OK);
}

/*************************************************************************
 * Function Name: USB_EpRead
 * Parameters: USB_Endpoint_t EndPoint, pInt32U pData, pInt32U pCount
 *
 * Return: USB_ErrorCodes_t
 *
 * Description: Endpoint Read (OUT)
 *
 *************************************************************************/
USB_ErrorCodes_t USB_EpRead (USB_Endpoint_t EndPoint, pInt32U pData,
                             pInt32U pCount)
{
pInt32U pDst;
Int32U Count = *pCount, EpCount;
EpState_t EpState;
  if(EpInfo[EndPoint].pEpCtrl == NULL)
  {
    return(USB_EP_NOT_VALID);
  }
  else
  {
    EpState = (EpState_t) EpCtrlGet_STAT_RX(EpInfo[EndPoint].pEpCtrl);
    if(EpState == EP_DISABLED)
    {
      return(USB_EP_FATAL_ERROR);
    }
  }
  switch (EpCtrlGet_EP_TYPE(EpInfo[EndPoint].pEpCtrl))
  {
  case EP_CTRL:
    if(EpState == EP_VALID)
    {
      return(UB_EP_SETUP_UNDERRUN);
    }
    EpCount = ReadEpDTB_CountRx(EpInfo[EndPoint].EpSlot) & 0x3FF;
    *pCount = MIN(EpCount,Count);
    pDst = (pInt32U)(ReadEpDTB_AddrRx(EpInfo[EndPoint].EpSlot)+\
            (Int32U)__segment_begin("USB_PACKET_MEMORY"));
    for(Int32U i = 0; i < *pCount; i += 4 )
    {
      *pData++=*pDst++;
    }
    EpCtrlSet_STAT_RX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
    break;
  case EP_ISO:
    if(EpCtrlGet_DTOG_RX(EpInfo[EndPoint].pEpCtrl))
    {
      EpCount = ReadEpDTB_CountTx(EpInfo[EndPoint].EpSlot) & 0x3FF;
      pDst = (pInt32U)(ReadEpDTB_AddrTx(EpInfo[EndPoint].EpSlot)+\
              (Int32U)__segment_begin("USB_PACKET_MEMORY"));

    }
    else
    {
      EpCount = ReadEpDTB_CountRx(EpInfo[EndPoint].EpSlot) & 0x3FF;
      pDst = (pInt32U)(ReadEpDTB_AddrRx(EpInfo[EndPoint].EpSlot)+\
              (Int32U)__segment_begin("USB_PACKET_MEMORY"));
    }
    *pCount = MIN(EpCount,Count);
    for(Int32U i = 0; i < *pCount; i += 4 )
    {
      *pData++=*pDst++;
    }
    break;
  case EP_INTERRUPT:
    if (EpState == EP_STALL)
    {
      return(USB_EP_STALLED);
    }
    else if (EpState == EP_VALID)
    {
      return(UB_EP_SETUP_UNDERRUN);
    }
    EpCount = ReadEpDTB_CountRx(EpInfo[EndPoint].EpSlot) & 0x3FF;
    *pCount = MIN(EpCount,Count);
    pDst = (pInt32U)(ReadEpDTB_AddrRx(EpInfo[EndPoint].EpSlot)+\
            (Int32U)__segment_begin("USB_PACKET_MEMORY"));
    for(Int32U i = 0; i < *pCount; i += 4 )
    {
      *pData++=*pDst++;
    }
    EpCtrlSet_STAT_RX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
    break;
  default:  // BULK
    if(EpCtrlGet_EP_KIND(EpInfo[EndPoint].pEpCtrl))
    {
      // Double buffered
      if(!EpCtrlGet_DTOG_TX(EpInfo[EndPoint].pEpCtrl))
      {
        EpCount = ReadEpDTB_CountRx(EpInfo[EndPoint].EpSlot) & 0x3FF;
        pDst = (pInt32U)(ReadEpDTB_AddrRx(EpInfo[EndPoint].EpSlot)+\
                (Int32U)__segment_begin("USB_PACKET_MEMORY"));
      }
      else
      {
        EpCount = ReadEpDTB_CountTx(EpInfo[EndPoint].EpSlot) & 0x3FF;
        pDst = (pInt32U)(ReadEpDTB_AddrTx(EpInfo[EndPoint].EpSlot)+\
                (Int32U)__segment_begin("USB_PACKET_MEMORY"));
      }
      *pCount = MIN(EpCount,Count);
      for(Int32U i = 0; i < *pCount; i += 4 )
      {
        *pData++=*pDst++;
      }
      EpCtrlToggle_DTOG_TX(EpInfo[EndPoint].pEpCtrl);
      if (EpCtrlGet_STAT_RX(EpInfo[EndPoint].pEpCtrl) == EP_NAK)
      {
        EpCtrlSet_STAT_RX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
      }
    }
    else
    {
      // Single buffer
      if (EpState == EP_STALL)
      {
        return(USB_EP_STALLED);
      }
      else if (EpState == EP_VALID)
      {
        return(UB_EP_SETUP_UNDERRUN);
      }
      EpCount = ReadEpDTB_CountRx(EpInfo[EndPoint].EpSlot) & 0x3FF;
      *pCount = MIN(EpCount,Count);
      pDst = (pInt32U)(ReadEpDTB_AddrRx(EpInfo[EndPoint].EpSlot)+\
              (Int32U)__segment_begin("USB_PACKET_MEMORY"));
      for(Int32U i = 0; i < *pCount; i += 4 )
      {
        *pData++=*pDst++;
      }
      EpCtrlSet_STAT_RX(EpInfo[EndPoint].pEpCtrl,EP_VALID);
    }
  }
  return (USB_OK);
}

/*************************************************************************
 * Function Name: USB_ISR
 * Parameters: none
 *
 * Return: none
 *
 * Description: USB interrupt subroutine
 *
 *************************************************************************/
#if USB_HIGH_PRIORITY_EVENT > 0
void USB_HIGH_ISR (void)
{
IntrStatus_t IntrStatus;
Int32U EpPhAddr;
  IntrStatus.Status  = USB_ISTR;
  IntrStatus.Status &= USB_CNTR | 0x1F;
  if(IntrStatus.CTR)
  {
    EpPhAddr = (IntrStatus.EP_ID << 1) + ((IntrStatus.DIR)?0:1);
    if(IntrStatus.DIR == 1)
    {
      EpPhAddr = IntrStatus.EP_ID << 1;
      EpCtrlClr_CTR_RX(EpInfo[EpPhAddr].pEpCtrl);
      if(UsbUserFun[EpPhAddr] != NULL)
      {
        UsbUserFun[EpPhAddr]((void *)UsbDataOutPacket);
      }
    }
    else
    {
      EpPhAddr = (IntrStatus.EP_ID << 1) + 1;
      EpCtrlClr_CTR_TX(EpInfo[EpPhAddr].pEpCtrl);
      if(UsbUserFun[EpPhAddr] != NULL)
      {
        UsbUserFun[EpPhAddr]((void *)UsbDataInPacket);
      }
    }
  }
}
#endif
	
/*************************************************************************
 * Function Name: USB_ISR
 * Parameters: none
 *
 * Return: none
 *
 * Description: USB interrupt subroutine
 *
 *************************************************************************/
void USB_ISR (void)
{
IntrStatus_t IntrStatus;
Int32U EpPhAddr;
Boolean SetupPacket;
  IntrStatus.Status  = USB_ISTR;
  IntrStatus.Status &= USB_CNTR | 0x1F;

#if USB_DOVR_EVENT > 0
  // DMA over / underrun (for speed up retry process)
  if(IntrStatus.DOVR)
  {
    USB_ISTR = ~bmDOVR;
    if(UsbUserFun[UsbDovrEvent] != NULL)
    {
      UsbUserFun[UsbDovrEvent]((void *)Val);
    }
  }
#endif

#if USB_ERROR_EVENT > 0
  // USB engine error interrupt
  if(IntrStatus.ERR)
  {
    USB_ISTR = ~bmERRM;
    if(UsbUserFun[UsbErrorEvent] != NULL)
    {
      UsbUserFun[UsbErrorEvent]((void *)Val);
    }
  }
#endif

  // Device reset
  if(IntrStatus.RESET)
  {
    USB_ISTR = ~bmRESETM;
    USB_HwReset();
    if(UsbUserFun[UsbResetEvent] != NULL)
    {
        UsbUserFun[UsbResetEvent](NULL);
    }
    if(UsbUserFun[UsbSuspendEvent] != NULL)
    {
      UsbUserFun[UsbSuspendEvent]((void *)0);
    }
  }

  // Frame interrupt
#if USB_SOF_EVENT > 0
  if(IntrStatus.SOF)
  {
    USB_ISTR = ~bmSOFM;
    if(UsbUserFun[UsbSofEvent] != NULL)
    {
      UsbUserFun[UsbSofEvent]((void *)NULL);
    }
  }
#endif

#if USB_WKUP_EVENT
  // Wake-Up interrupt
  if(IntrStatus.WKUP)
  {
    USB_ISTR = ~bmWKUPM;
    USB_Resume(USB_RESUME_WAKE_UP_EVENT);
    if(UsbUserFun[UsbWakeUpEvent] != NULL)
    {
      UsbUserFun[UsbWakeUpEvent]((void *)Val);
    }
  }
#endif

  // Device suspend
  if(IntrStatus.SUSP)
  {
    USB_ISTR = ~bmSUSPM;
    if(UsbUserFun[UsbSuspendEvent] != NULL)
    {
      UsbUserFun[UsbSuspendEvent]((void *)1);
    }
  }

  // Expexted Frame interrupt
  if(IntrStatus.ESOF)
  {
    USB_ISTR = ~bmESOFM;
    USB_Resume(USB_RESUME_SOF_EVENT);
  }

  // Low priority EP interrupt
  if(IntrStatus.CTR)
  {
    EpPhAddr = (IntrStatus.EP_ID << 1) + (IntrStatus.DIR?0:1);
    if(EpPhAddr == CTRL_ENP_OUT)
    {
      SetupPacket = EpCtrlGet_SETUP(EpInfo[EpPhAddr].pEpCtrl);
      EpCtrlClr_CTR_RX(EpInfo[EpPhAddr].pEpCtrl);
      if(UsbUserFun[UsbEp0Out] != NULL)
      {
        if(SetupPacket)
        {
          UsbUserFun[UsbEp0Out]((void *)UsbSetupPacket);
        }
        else
        {
          UsbUserFun[UsbEp0Out]((void *)UsbDataOutPacket);
        }
      }
    }
    else if (EpPhAddr & 1)
    {
      EpCtrlClr_CTR_TX(EpInfo[EpPhAddr].pEpCtrl);
      if(UsbUserFun[EpPhAddr] != NULL)
      {
        UsbUserFun[EpPhAddr]((void *)UsbDataInPacket);
      }
    }
    else
    {
      EpCtrlClr_CTR_RX(EpInfo[EpPhAddr].pEpCtrl);
      if(UsbUserFun[EpPhAddr] != NULL)
      {
        UsbUserFun[EpPhAddr]((void *)UsbDataOutPacket);
      }
    }
  }
}

/*************************************************************************
 * Function Name: USB_EpLogToPhysAdd
 * Parameters: Int8U EpLogAdd
 *
 * Return: USB_Endpoint_t
 *
 * Description: Convert the logical to physical address
 *
 *************************************************************************/
USB_Endpoint_t USB_EpLogToPhysAdd (Int8U EpLogAdd)
{
USB_Endpoint_t Address = (USB_Endpoint_t)((EpLogAdd & 0x0F)<<1);
  if(EpLogAdd & 0x80)
  {
    ++Address;
  }
  return(Address);
}

/*************************************************************************
 * Function Name: USB_GetFrameNumb
 * Parameters: none
 *
 * Return: Int32U
 *
 * Description: Retunr curent value of SOF number
 *
 *************************************************************************/
#if USB_SOF_EVENT > 0
Int32U USB_GetFrameNumb (void)
{
  return(USB_FNR_bit.FN);
}
#endif

/*************************************************************************
 *                          U S B   D M A  P a r t                       *
 *************************************************************************/
#if USB_DMA > 0
Lli_t UsbRxDmaLli[USB_DMA_RX_MAX];
Lli_t UsbTxDmaLli[USB_DMA_TX_MAX];

/*************************************************************************
 * Function Name: USB_DmaReset
 * Parameters:  none
 *
 * Return: none
 *
 * Description: Reset USB DMA
 *
 *************************************************************************/
void USB_DmaReset (void)
{
  // Disable USB DMA
  USB_DMACR1 = \
  USB_DMACR2 = \
  USB_DMACR3 = 0;
}

/*************************************************************************
 * Function Name: USB_DmaTxHandler
 * Parameters:  void * Arg
 *
 * Return: void *
 *
 * Description: USB DMA Tx interrupt handler
 *
 *************************************************************************/
void * USB_DmaTxHandler (void * Arg)
{
  if(UsbDmaUserFun[UsbDmaUnLliTx] != NULL)
  {
    return(UsbDmaUserFun[UsbDmaUnLliTx](Arg));
  }
  return(NULL);
}

/*************************************************************************
 * Function Name: USB_DmaRxHandler
 * Parameters:  void * Arg
 *
 * Return: void *
 *
 * Description: USB DMA Rx interrupt handler
 *
 *************************************************************************/
void * USB_DmaRxHandler (void * Arg)
{
  if(UsbDmaUserFun[UsbDmaUnLliRx] != NULL)
  {
    return(UsbDmaUserFun[UsbDmaUnLliRx](Arg));
  }
  return(NULL);
}

/*************************************************************************
 * Function Name: USB_DmaInitTransfer
 * Parameters: USB_Endpoint_t EndPoint, Int32U DmaDescInd,
 *             pInt32U pData, Int32U EpMaxSize, Int32U Size
 *             pDmaIsoPacket_t pDmaIsoPacket,  Boolean EpTransferType
 *
 * Return: UsbDmaStateCode_t
 *
 * Description: Init Transfer by DMA
 *
 *************************************************************************/
UsbDmaStateCode_t USB_DmaInitTransfer (USB_Endpoint_t EndPoint, DmaChannel_t DmaChannel,
                                       pInt32U pData, Int32U Size)
{
pEpInfo = EpInfo[EndPoint];
Int32U Count = 1;
pLli_t pLli;
pLli_t pLliPrev;
Int32U CurrBuff;
Cc_t CC = {0};

  // Is EP valid?
  if(pEpInfo->pEpCtrl == NULL)
  {
    return(UsbDmaParametersError);
  }
  if(DmaReq(DmaReqEnableChannels) & (1UL << DmaChannel))
  {
    return(UsbDmaBeingServiced);
  }

  // DMA transfer with LLI
  if(EndPoint & 1)
  {
    pLli = UsbTxDmaLli;
    pLliPrev = NULL;
    CurrBuff

    CC.TransferSize = pEpInfo->MaxSize;
    CC.DI = CC.SI = 1;          // Source and destination Addr incremet
    CC.Swidth = CC.Dwidth = 2;  // 32 bit
    if (pEpInfo->MaxSize >= 256)
    {
      CC.SBSize = CC.DBSize = 7;  // 256 bytes
    }
    else if (pEpInfo->MaxSize >= 128)
    {
      CC.SBSize = CC.DBSize = 6;  // 128 bytes
    }
    else if (pEpInfo->MaxSize >= 64)
    {
      CC.SBSize = CC.DBSize = 5;  // 64 bytes
    }
    else if (pEpInfo->MaxSize >= 32)
    {
      CC.SBSize = CC.DBSize = 4;  // 32 bytes
    }
    else if (pEpInfo->MaxSize >= 16)
    {
      CC.SBSize = CC.DBSize = 3;  // 16 bytes
    }
    else if (pEpInfo->MaxSize >= 8)
    {
      CC.SBSize = CC.DBSize = 2;  // 8 bytes
    }
    else if (pEpInfo->MaxSize >= 4)
    {
      CC.SBSize = CC.DBSize = 1;  // 4 bytes
    }
    USB_DMABSIZE_bit.DBSIZE = CC.DBSize;
    while(Size > pEpInfo->MaxSize)
    {
      pLli->SrcAdd  = pData;
      pLli->DestAdd = (Int32U)((CurrBuff&1)?pEpInfo->AddrBuf2:pEpInfo->AddrBuf1);
      pLli->pLli    = pLliPrev;
      pLli->CC.Data = CC_Hold;
      pLliPrev      = pLli++;
      Size         -= pEpInfo->MaxSize;
      pData        += pEpInfo->MaxSize;
      ++CurrBuff;
      ++Count;
    }
    pLli->SrcAdd  = pData;
    pLli->DestAdd = (Int32U)((CurrBuff&1)?pEpInfo->AddrBuf2:pEpInfo->AddrBuf1);
    pLli->pLli    = NULL;
    pLli->CC.Data = CC_Hold;
    pLli->CC.TransferSize = Size;
    pLli->CC.TCIE = 1;
  }
  else
  {
    if (pEpInfo->MaxSize >= 256)
    {
      CC.SBSize = CC.DBSize = 7;  // 256 bytes
    }
    else if (pEpInfo->MaxSize >= 128)
    {
      CC.SBSize = CC.DBSize = 6;  // 128 bytes
    }
    else if (pEpInfo->MaxSize >= 64)
    {
      CC.SBSize = CC.DBSize = 5;  // 64 bytes
    }
    else if (pEpInfo->MaxSize >= 32)
    {
      CC.SBSize = CC.DBSize = 4;  // 32 bytes
    }
    else if (pEpInfo->MaxSize >= 16)
    {
      CC.SBSize = CC.DBSize = 3;  // 16 bytes
    }
    else if (pEpInfo->MaxSize >= 8)
    {
      CC.SBSize = CC.DBSize = 2;  // 8 bytes
    }
    else if (pEpInfo->MaxSize >= 4)
    {
      CC.SBSize = CC.DBSize = 1;  // 4 bytes
    }
  }

/*************************************************************************
 * Function Name: USB_DmaInitTransfer
 * Parameters: USB_Endpoint_t EndPoint, Int32U DmaDescInd,
 *             pInt32U pData, Int32U EpMaxSize, Int32U Size
 *             pDmaIsoPacket_t pDmaIsoPacket,  Boolean EpTransferType
 *
 * Return: UsbDmaStateCode_t
 *
 * Description: Init Transfer by DMA
 *
 *************************************************************************/
void  USB_DmaCreateLli(pLli_t pLli, pEpInfo_t pEpInfo,
                         Boolean Src_Dst, Boolean CurrentFreeBuffer,
                         pInt8U pBuffer, Int32U Size);
{
Int32U Temp;
  if(pEpInfo->MaxSize)
}


    else
  {
    // DMA transfer without LLI
    if(EpInfo[EndPoint].MaxSize < Size)
    {
      return (UsbDmaParametersError);
    }
    if(EndPoint & 1)
    {
      USB_DMACR3_bit.LK_TX_EN = 0;
      USB_DMACR2 = EpInfo[EndPoint].EpSlot | (1UL << 12);
      if(DmaChannelInit(DmaChannel,UsbTx,UsbTx,Mem2PerPer,FALSE) != DmaPass)
      {
        return(UsbDmaBeingServiced);
      }
      DMA_UserFuncRegistered(USB_DmaUnLliTxHandler,DmaChannel);
    }
    else
    {
      USB_DMACR3_bit.LK_RX_EN = 0;
      USB_DMACR1 = (1UL << EpInfo[EndPoint].EpSlot);
      if(DmaChannelInit(DmaChannel,UsbRx,UsbRx,Per2MemPer,FALSE) != DmaPass)
      {
        return(UsbDmaBeingServiced);
      }
      DMA_UserFuncRegistered(USB_DmaUnLliRxHandler,DmaChannel);
    }
  }
  return(UsbDmaPass);
}

#endif // USB_DMA > 0
